"""Utilities module."""
import asyncio
import logging
from typing import Callable
from warnings import warn

_LOGGER = logging.getLogger(__name__)


def first_or_none(lst, predicate):
    filter_iter = filter(predicate, lst)
    return next(filter_iter, None)


def add_async_job(target: Callable, *args):
    loop = asyncio.get_event_loop()

    if asyncio.iscoroutine(target):
        task = loop.create_task(target)
    elif asyncio.iscoroutinefunction(target):
        task = loop.create_task(target(*args))
    else:
        task = loop.run_in_executor(None, target, *args)

    return task


def send_deprecation_warning(old_name, new_name):
    message = f"{old_name} has been deprecated in favor of {new_name}, the alias will be removed in the future"
    warn(
        message,
        DeprecationWarning,
        stacklevel=2,
    )
    _LOGGER.warn(message)



"""Module that implements the Vivint class."""
import logging
from typing import List, Optional

import aiohttp
from aiohttp.client_exceptions import ClientConnectionError

import pubnub
from pubnub.enums import PNHeartbeatNotificationOptions, PNReconnectionPolicy
from pubnub.pnconfiguration import PNConfiguration

from .const import (
    AuthUserAttribute,
    PubNubMessageAttribute,
    SystemAttribute,
    UserAttribute,
)
from .pubnub import (
    PN_CHANNEL,
    PN_SUBSCRIBE_KEY,
    VivintPubNubSubscribeListener,
    pubnub_asyncio,
)
from .system import System
from .utils import first_or_none
from .vivintskyapi import VivintSkyApi

_LOGGER = logging.getLogger(__name__)


class Account:
    """Class for interacting with VivintSky API using asyncio"""

    def __init__(
        self,
        username: str,
        password: str,
        client_session: Optional[aiohttp.ClientSession] = None,
    ):
        self.__connected = False
        self.__pubnub: pubnub_asyncio.PubNubAsyncio = None
        self.__pubnub_listener: VivintPubNubSubscribeListener = None
        self.vivintskyapi = VivintSkyApi(username, password, client_session)
        self.systems: List[System] = []

    @property
    def connected(self):
        """Return True if connected."""
        return self.__connected

    async def connect(
        self,
        load_devices: bool = False,
        subscribe_for_realtime_updates: bool = False,
    ) -> None:
        """Connects to vivintsky cloud service."""

        _LOGGER.debug("connecting to vivintsky")
        # initialize the vivintsky cloud session
        authuser_data = await self.vivintskyapi.connect()

        # load all systems, panels and devices
        if load_devices:
            _LOGGER.debug("loading devices")
            await self.refresh(authuser_data)

        # subscribe to pubnub for realtime updates
        if subscribe_for_realtime_updates:
            _LOGGER.debug("subscribing to pubnub for realtime updates")
            await self.subscribe_for_realtime_updates(authuser_data)

        self.__connected = True

    async def disconnect(self) -> None:
        _LOGGER.debug("disconnecting from vivintsky")
        if self.connected:
            if self.__pubnub:
                self.__pubnub.remove_listener(self.__pubnub_listener)
                self.__pubnub.unsubscribe_all()
                self.__pubnub.stop()
            await self.vivintskyapi.disconnect()
        self.__connected = False

    async def refresh(self, authuser_data: dict = None) -> None:
        # make a call to vivint's authuser endpoint to get a list of all the system_accounts (locations) & panels if not supplied
        if not authuser_data:
            try:
                authuser_data = await self.vivintskyapi.get_authuser_data()
            except ClientConnectionError:
                _LOGGER.error("Unable to refresh system(s).")

        if authuser_data:
            # for each system_account, make another call to load all the devices
            for system_data in authuser_data[AuthUserAttribute.USERS][
                UserAttribute.SYSTEM
            ]:
                # is this an existing account_system?
                system = first_or_none(
                    self.systems,
                    lambda system: system.id == system_data[SystemAttribute.PANEL_ID],
                )
                if system:
                    await system.refresh()
                else:
                    full_system_data = await self.vivintskyapi.get_system_data(
                        system_data[SystemAttribute.PANEL_ID]
                    )
                    self.systems.append(
                        System(
                            system_data.get(SystemAttribute.SYSTEM_NICKNAME),
                            full_system_data,
                            self.vivintskyapi,
                        )
                    )

            _LOGGER.debug(
                f"Refreshed {len(authuser_data[AuthUserAttribute.USERS][UserAttribute.SYSTEM])} system(s)."
            )

    async def subscribe_for_realtime_updates(self, authuser_data: dict = None) -> None:
        """Subscribes to PubNub for realtime updates."""
        # make a call to vivint's authuser endpoint to get message broadcast channel if not supplied
        if not authuser_data:
            authuser_data = await self.vivintskyapi.get_authuser_data()

        pubnub.set_stream_logger("pubnub", logging.INFO)

        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = PN_SUBSCRIBE_KEY
        pnconfig.ssl = True
        pnconfig.reconnect_policy = PNReconnectionPolicy.LINEAR
        pnconfig.heartbeat_notification_options = (
            PNHeartbeatNotificationOptions.FAILURES
        )

        self.__pubnub = pubnub_asyncio.PubNubAsyncio(pnconfig)
        self.__pubnub_listener = VivintPubNubSubscribeListener(
            self.handle_pubnub_message
        )
        self.__pubnub.add_listener(self.__pubnub_listener)

        pn_channel = f"{PN_CHANNEL}#{authuser_data[AuthUserAttribute.USERS][UserAttribute.MESSAGE_BROADCAST_CHANNEL]}"
        self.__pubnub.subscribe().channels(pn_channel).with_presence().execute()

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        panel_id = message.get(PubNubMessageAttribute.PANEL_ID)
        if not panel_id:
            _LOGGER.info("No panel id specified - ignoring pubnub message.")
            return

        system = first_or_none(self.systems, lambda system: system.id == panel_id)
        if not system:
            _LOGGER.info(f"No system found with id {panel_id}.")
            return

        system.handle_pubnub_message(message)



"""This package contains the various devices attached to a Vivint system."""
from __future__ import annotations

import asyncio
import concurrent.futures
from typing import TYPE_CHECKING, Callable, Dict, List, Optional

from ..const import VivintDeviceAttribute as Attribute
from ..entity import Entity
from ..enums import CapabilityCategoryType, CapabilityType
from ..vivintskyapi import VivintSkyApi

if TYPE_CHECKING:
    from .alarm_panel import AlarmPanel


def get_device_class(device_type: str) -> Callable:
    """Maps a device_type string to the class that implements that device."""
    from ..enums import DeviceType
    from . import UnknownDevice
    from .camera import Camera
    from .door_lock import DoorLock
    from .garage_door import GarageDoor
    from .switch import BinarySwitch, MultilevelSwitch
    from .thermostat import Thermostat
    from .wireless_sensor import WirelessSensor

    mapping = {
        DeviceType.BINARY_SWITCH: BinarySwitch,
        DeviceType.CAMERA: Camera,
        DeviceType.DOOR_LOCK: DoorLock,
        DeviceType.GARAGE_DOOR: GarageDoor,
        DeviceType.MULTILEVEL_SWITCH: MultilevelSwitch,
        DeviceType.THERMOSTAT: Thermostat,
        DeviceType.WIRELESS_SENSOR: WirelessSensor,
    }

    return mapping.get(DeviceType(device_type), UnknownDevice)


class VivintDevice(Entity):
    """Class to implement a generic vivint device."""

    def __init__(self, data: dict, alarm_panel: AlarmPanel = None):
        super().__init__(data)
        self.alarm_panel = alarm_panel
        self._manufacturer = None
        self._model = None
        self._capabilities = (
            {
                CapabilityCategoryType(capability_category.get(Attribute.TYPE)): [
                    CapabilityType(capability)
                    for capability in capability_category.get(Attribute.CAPABILITY)
                ]
                for capability_category in data.get(Attribute.CAPABILITY_CATEGORY)
            }
            if data.get(Attribute.CAPABILITY_CATEGORY)
            else None
        )

    def __repr__(self):
        """Custom repr method"""
        return f"<{self.__class__.__name__} {self.id}, {self.name}>"

    @property
    def id(self) -> int:
        """Device's id."""
        return self.data[Attribute.ID]

    @property
    def name(self) -> Optional[str]:
        """Device's name."""
        return self.data.get(Attribute.NAME)

    @property
    def capabilities(
        self,
    ) -> Optional[Dict[CapabilityCategoryType, List[CapabilityType]]]:
        """Device capabilities."""
        return self._capabilities

    @property
    def manufacturer(self):
        """Return the manufacturer for this device."""
        if not self._manufacturer and self.data.get("zpd"):
            self.get_zwave_details()
        return self._manufacturer

    @property
    def model(self):
        """Return the model for this device."""
        if not self._model and self.data.get("zpd"):
            self.get_zwave_details()
        return self._model

    @property
    def serial_number(self) -> str:
        """Return the serial number for this device."""
        serial_number = self.data.get(Attribute.SERIAL_NUMBER_32_BIT)
        serial_number = (
            serial_number if serial_number else self.data.get(Attribute.SERIAL_NUMBER)
        )
        return serial_number

    @property
    def software_version(self) -> str:
        """Return the software version of this device, if any."""
        # panels
        current_software_version = self.data.get(Attribute.CURRENT_SOFTWARE_VERSION)
        # z-wave devices (some)
        firmware_version = (
            ".".join(
                [
                    str(i)
                    for s in self.data.get(Attribute.FIRMWARE_VERSION) or []
                    for i in s
                ]
            )
            or None
        )
        return current_software_version or firmware_version

    @property
    def vivintskyapi(self) -> VivintSkyApi:
        """Instance of VivintSkyApi."""
        assert self.alarm_panel, """no alarm panel set for this device"""
        return self.alarm_panel.system.vivintskyapi

    def get_zwave_details(self):
        if self.data.get("zpd") is None:
            return None

        pool = concurrent.futures.ThreadPoolExecutor()
        result = pool.submit(asyncio.run, self.get_zwave_details_async()).result()
        return result

    async def get_zwave_details_async(self):
        manufacturer_id = f"{self.data.get('manid'):04x}"
        product_id = f"{self.data.get('prid'):04x}"
        product_type_id = f"{self.data.get('prtid'):04x}"
        result = await self.vivintskyapi.get_zwave_details(
            manufacturer_id, product_id, product_type_id
        )
        [self._manufacturer, self._model] = result
        return result


class UnknownDevice(VivintDevice):
    """Describe an unknown/unsupported vivint device."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}|{self.data[Attribute.TYPE]} {self.id}>"



"""File for various patches to PubNub until fixed in module."""
from pubnub import pubnub, pubnub_asyncio, utils
from pubnub.endpoints.presence.heartbeat import Heartbeat
from pubnub.endpoints.pubsub.subscribe import Subscribe
from pubnub.enums import (
    PNHeartbeatNotificationOptions,
    PNOperationType,
    PNReconnectionPolicy,
    PNStatusCategory,
)
from pubnub.errors import (
    PNERR_CLIENT_TIMEOUT,
    PNERR_CONNECTION_ERROR,
    PNERR_REQUEST_CANCELLED,
)
from pubnub.exceptions import PubNubException
from pubnub.pubnub_asyncio import PubNubAsyncioException, aiohttp, asyncio, logger


async def patched_request_future(self, options_func, cancellation_event):
    """
    Patched `request_future` function on pubnub's `PubNubAsyncio` class.

    Return status of PNStatusCategory.PNNetworkIssuesCategory when ClientConnectorError occurs.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    try:
        res = await self._request_helper(options_func, cancellation_event)
        return res
    except PubNubException as e:
        return PubNubAsyncioException(result=None, status=e.status)
    except asyncio.TimeoutError:
        return PubNubAsyncioException(
            result=None,
            status=options_func().create_status(
                PNStatusCategory.PNTimeoutCategory,
                None,
                None,
                exception=PubNubException(pn_error=PNERR_CLIENT_TIMEOUT),
            ),
        )
    except asyncio.CancelledError:
        return PubNubAsyncioException(
            result=None,
            status=options_func().create_status(
                PNStatusCategory.PNCancelledCategory,
                None,
                None,
                exception=PubNubException(pn_error=PNERR_REQUEST_CANCELLED),
            ),
        )
    except aiohttp.client_exceptions.ClientConnectorError:
        return PubNubAsyncioException(
            result=None,
            status=options_func().create_status(
                PNStatusCategory.PNNetworkIssuesCategory,
                None,
                None,
                exception=PubNubException(pn_error=PNERR_CONNECTION_ERROR),
            ),
        )
    except Exception as e:
        return PubNubAsyncioException(
            result=None,
            status=options_func().create_status(
                PNStatusCategory.PNUnknownCategory, None, None, e
            ),
        )


pubnub_asyncio.PubNubAsyncio.request_future = patched_request_future


async def patched_register_heartbeat_timer(self):
    """
    Patched `_register_heartbeat_timer` function on pubnub's `AsyncioReconnectionManager` class.

    Raise exception when there is an error on heartbeat call so that reconnect logic actually happens.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    while True:
        self._recalculate_interval()

        await asyncio.sleep(self._timer_interval)

        logger.debug("reconnect loop at: %s" % utils.datetime_now())

        try:
            result = await self._pubnub.time().future()
            if result.status.is_error():
                raise result.status.error_data.exception
            self._connection_errors = 1
            self._callback.on_reconnect()
            break
        except Exception:
            if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.EXPONENTIAL:
                logger.debug(
                    "reconnect interval increment at: %s" % utils.datetime_now()
                )
                self._connection_errors += 1


pubnub_asyncio.AsyncioReconnectionManager._register_heartbeat_timer = (
    patched_register_heartbeat_timer
)


async def patched_start_subscribe_loop(self):
    """
    Patched `_start_subscribe_loop` function on pubnub's `AsyncioSubscriptionManager` class.

    Correctly call `_start_subscribe_loop` on timeouts.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    self._stop_subscribe_loop()

    await self._subscription_lock.acquire()

    combined_channels = self._subscription_state.prepare_channel_list(True)
    combined_groups = self._subscription_state.prepare_channel_group_list(True)

    if len(combined_channels) == 0 and len(combined_groups) == 0:
        self._subscription_lock.release()
        return

    self._subscribe_request_task = asyncio.ensure_future(
        Subscribe(self._pubnub)
        .channels(combined_channels)
        .channel_groups(combined_groups)
        .timetoken(self._timetoken)
        .region(self._region)
        .filter_expression(self._pubnub.config.filter_expression)
        .future()
    )

    e = await self._subscribe_request_task

    if self._subscribe_request_task.cancelled():
        self._subscription_lock.release()
        return

    if e.is_error():
        if (
            e.status is not None
            and e.status.category == PNStatusCategory.PNCancelledCategory
        ):
            self._subscription_lock.release()
            return

        if (
            e.status is not None
            and e.status.category == PNStatusCategory.PNTimeoutCategory
        ):
            asyncio.create_task(self._start_subscribe_loop())
            self._subscription_lock.release()
            return

        logger.error("Exception in subscribe loop: %s" % str(e))

        if (
            e.status is not None
            and e.status.category == PNStatusCategory.PNAccessDeniedCategory
        ):
            e.status.operation = PNOperationType.PNUnsubscribeOperation

        # TODO: raise error
        self._listener_manager.announce_status(e.status)

        self._reconnection_manager.start_polling()
        self._subscription_lock.release()
        self.disconnect()
        return
    else:
        self._handle_endpoint_call(e.result, e.status)
        self._subscription_lock.release()
        self._subscribe_loop_task = asyncio.ensure_future(self._start_subscribe_loop())

    self._subscription_lock.release()


pubnub_asyncio.AsyncioSubscriptionManager._start_subscribe_loop = (
    patched_start_subscribe_loop
)


async def patched_perform_heartbeat_loop(self):
    """
    Patched `_perform_heartbeat_loop` function on pubnub's `AsyncioSubscriptionManager` class.

    Correctly adheres to setting for heartbeat_notification_options.
    Call `_start_subscribe_loop` on timeouts and network issues.

    See https://github.com/pubnub/python/pull/101 for when this can be removed.
    """
    if self._heartbeat_call is not None:
        # TODO: cancel call
        pass

    cancellation_event = asyncio.Event()
    state_payload = self._subscription_state.state_payload()
    presence_channels = self._subscription_state.prepare_channel_list(False)
    presence_groups = self._subscription_state.prepare_channel_group_list(False)

    if len(presence_channels) == 0 and len(presence_groups) == 0:
        return

    try:
        heartbeat_call = (
            Heartbeat(self._pubnub)
            .channels(presence_channels)
            .channel_groups(presence_groups)
            .state(state_payload)
            .cancellation_event(cancellation_event)
            .future()
        )

        envelope = await heartbeat_call

        heartbeat_verbosity = self._pubnub.config.heartbeat_notification_options
        if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL or (
            envelope.status.is_error()
            and heartbeat_verbosity == PNHeartbeatNotificationOptions.FAILURES
        ):
            self._listener_manager.announce_status(envelope.status)

        if envelope.status is not None and envelope.status.category in [
            PNStatusCategory.PNTimeoutCategory,
            PNStatusCategory.PNNetworkIssuesCategory,
        ]:
            await self._start_subscribe_loop()

    except PubNubAsyncioException:
        pass

    finally:
        cancellation_event.set()


pubnub_asyncio.AsyncioSubscriptionManager._perform_heartbeat_loop = (
    patched_perform_heartbeat_loop
)


old_endpoint_name_for_operation = pubnub.TelemetryManager.endpoint_name_for_operation


@staticmethod
def patched_endpoint_name_for_operation(operation_type):
    """
    Patched `endpoint_name_for_operation` function on pubnub's `TelemetryManager` class.

    Avoids KeyError(6) when operation_type == PNOperationType.PNHeartbeatOperation.

    See https://github.com/pubnub/python/pull/100 for when this can be removed.
    """
    return (
        "hb"
        if operation_type == PNOperationType.PNHeartbeatOperation
        else old_endpoint_name_for_operation(operation_type)
    )


pubnub.TelemetryManager.endpoint_name_for_operation = (
    patched_endpoint_name_for_operation
)



"""Module that implements the Camera class."""
from datetime import datetime
from typing import Callable, List

from ..const import CameraAttribute as Attribute
from ..const import PanelCredentialAttribute
from . import VivintDevice
from .alarm_panel import AlarmPanel

# Some Vivint supported cameras may be connected directly to your local network
# and the Vivint API reports these as having direct access availiable (cda).
# Ping cameras (alpha_cs6022_camera_device) can be setup to connect to your own
# Wi-Fi via a complicated process involving removing and resetting the camera,
# initiating WPS on your Wi-Fi and the Ping camera, and then adding a new camera
# from the panel within a limited timeframe. The Ping camera, however, seems to
# still have a VPN connection with the panel that prevents direct access except
# on very rare occasions. As such, we want to skip this model from direct access.
SKIP_DIRECT = ["alpha_cs6022_camera_device"]


class Camera(VivintDevice):
    """Represents a vivint camera device."""

    def __init__(self, data: dict, alarm_panel: AlarmPanel):
        super().__init__(data, alarm_panel)
        self.__thumbnail_ready_callbacks: List[Callable] = list()

    @property
    def manufacturer(self):
        """Return the camera manufacturer."""
        return self.data.get(Attribute.ACTUAL_TYPE).split("_")[0].title()

    @property
    def model(self):
        """Return the camera model."""
        return self.data.get(Attribute.ACTUAL_TYPE).split("_")[1].upper()

    @property
    def serial_number(self) -> str:
        """Return the camera's mac address as the serial number."""
        return self.mac_address

    @property
    def software_version(self) -> str:
        """Return the camera's software version."""
        return self.data.get(Attribute.SOFTWARE_VERSION)

    @property
    def capture_clip_on_motion(self) -> bool:
        "Return True if capture clip on motion is active."
        return self.data[Attribute.CAPTURE_CLIP_ON_MOTION]

    @property
    def ip_address(self) -> str:
        "Camera's IP address."
        return self.data[Attribute.CAMERA_IP_ADDRESS]

    @property
    def mac_address(self) -> str:
        """Camera's MAC Address"""
        return self.data[Attribute.CAMERA_MAC]

    @property
    def is_in_privacy_mode(self) -> bool:
        """Return True if privacy mode is active."""
        return self.data[Attribute.CAMERA_PRIVACY]

    @property
    def is_online(self) -> bool:
        """Return True if camera is online."""
        return self.data[Attribute.ONLINE]

    @property
    def wireless_signal_strength(self) -> int:
        """Camera's wireless signal strength."""
        return self.data[Attribute.WIRELESS_SIGNAL_STRENGTH]

    def add_thumbnail_ready_callback(self, callback: Callable) -> None:
        """Register a thumbnail_ready callback."""
        self.__thumbnail_ready_callbacks.append(callback)

    async def request_thumbnail(self) -> None:
        """Request a new thumbnail for the camera."""
        await self.vivintskyapi.request_camera_thumbnail(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id
        )

    async def get_thumbnail_url(self) -> str:
        """Returns the latest camera thumbnail URL."""
        # Sometimes this date field comes back with a "Z" at the end
        # and sometimes it doesn't, so let's just safely remove it.
        camera_thumbnail_date = datetime.strptime(
            self.data[Attribute.CAMERA_THUMBNAIL_DATE].replace("Z", ""),
            "%Y-%m-%dT%H:%M:%S.%f",
        )
        thumbnail_timestamp = int(camera_thumbnail_date.timestamp() * 1000)

        return await self.vivintskyapi.get_camera_thumbnail_url(
            self.alarm_panel.id,
            self.alarm_panel.partition_id,
            self.id,
            thumbnail_timestamp,
        )

    async def get_rtsp_url(self, internal: bool = False, hd: bool = False) -> str:
        """Returns the rtsp URL for the camera."""
        credentials = await self.alarm_panel.get_panel_credentials()
        url = self.data[f"c{'i' if internal else 'e'}u{'' if hd else 's'}"][0]
        return f"{url[:7]}{credentials[PanelCredentialAttribute.NAME]}:{credentials[PanelCredentialAttribute.PASSWORD]}@{url[7:]}"

    async def get_direct_rtsp_url(self, hd: bool = False) -> str:
        """Returns the direct rtsp url for this camera, in HD if requested, if any."""
        return (
            f"rtsp://{self.data[Attribute.USERNAME]}:{self.data[Attribute.PASSWORD]}@{self.ip_address}:{self.data[Attribute.CAMERA_IP_PORT]}/{self.data[Attribute.CAMERA_DIRECT_STREAM_PATH if hd else Attribute.CAMERA_DIRECT_STREAM_PATH_STANDARD]}"
            if self.data[Attribute.CAMERA_DIRECT_AVAILABLE]
            and self.data.get(Attribute.ACTUAL_TYPE) not in SKIP_DIRECT
            else None
        )

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message addressed to this camera."""
        super().handle_pubnub_message(message)

        if message.get(Attribute.CAMERA_THUMBNAIL_DATE):
            self._fire_callbacks(self.__thumbnail_ready_callbacks)



"""(deprecated) Module that defines various constants for interacting with the Vivint Sky API.

This module has been deprecated in favor of `const`.
"""


from .const import *
from .utils import send_deprecation_warning


def __getattr__(name):
    """Log a warning."""
    send_deprecation_warning("vivintpy.constants", "vivintpy.const")
    return name



"""Module that implements the DoorLock class."""
from ..const import ZWaveDeviceAttribute as Attributes
from . import VivintDevice


class DoorLock(VivintDevice):
    """Represents a vivint door lock device."""

    @property
    def battery_level(self) -> int:
        """Door lock's battery level."""
        return self.data[Attributes.BATTERY_LEVEL]

    @property
    def low_battery(self) -> bool:
        """Returns True if battery level is low."""
        return self.data[Attributes.LOW_BATTERY]

    @property
    def is_locked(self) -> bool:
        """Returns True if door lock is locked."""
        return self.data[Attributes.STATE]

    @property
    def node_online(self) -> bool:
        """Returns True if the node is online."""
        return self.data[Attributes.ONLINE]

    async def set_state(self, locked: bool) -> None:
        """Set door lock's state."""
        await self.vivintskyapi.set_lock_state(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, locked
        )

    async def lock(self) -> None:
        """Lock the door lock."""
        await self.set_state(True)

    async def unlock(self) -> None:
        """Unlock the door lock."""
        await self.set_state(False)



"""Module that implements the Entity class."""
import logging
from typing import Callable, List

from .utils import add_async_job

_LOGGER = logging.getLogger(__name__)


class Entity:
    """Describe a vivint entity."""

    def __init__(self, data: dict):
        self.__data = data
        self.__update_callbacks: Callable = list()

    @property
    def data(self) -> dict:
        """Return entity's raw data as returned by VivintSky API."""
        return self.__data

    def update_data(self, new_val: dict, override: bool = False) -> None:
        """Update entity's raw data."""
        if override:
            self.__data = new_val
        else:
            self.__data.update(new_val)

        self._fire_callbacks(self.__update_callbacks)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message directed to this entity."""
        self.update_data(message)

    def add_update_callback(self, callback: Callable) -> None:
        """Registers an update callback."""
        self.__update_callbacks.append(callback)

    def _fire_callbacks(self, callbacks: List[Callable], *args, **kwargs) -> None:
        """Execute callbacks.

        Handles both sync and async callbacks.
        """
        for callback in callbacks:
            try:
                add_async_job(callback)
            except Exception:
                _LOGGER.exception(
                    f"failed to execute callback for entity {self.__repr__()}"
                )



from .account import Account, Optional, aiohttp
from .utils import send_deprecation_warning


class Vivint(Account):
    """(deprecated) Vivint class (for backwards compatibility).

    This class has been deprecated in favor of `Account`.
    """

    def __init__(
        self,
        username: str,
        password: str,
        client_session: Optional[aiohttp.ClientSession] = None,
    ):
        send_deprecation_warning("vivintpy.vivint.Vivint", "vivintpy.account.Account")
        return super().__init__(username, password, client_session)



"""Module that implements the VivintPubNubSubscribeListener class."""
import logging
from enum import IntEnum, unique
from typing import Callable

from pubnub.callbacks import SubscribeCallback
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pubsub import PNMessageResult

from .pubnub_patches import pubnub_asyncio

PN_SUBSCRIBE_KEY = "sub-c-6fb03d68-6a78-11e2-ae8f-12313f022c90"
PN_CHANNEL = "PlatformChannel"

_LOGGER = logging.getLogger(__name__)


class VivintPubNubSubscribeListener(SubscribeCallback):
    """PubNub Subscribe Listener."""

    def __init__(self, message_received_callback: Callable):
        super().__init__()
        self.__message_received_callback = message_received_callback

    def status(
        self,
        pubnub: pubnub_asyncio.PubNubAsyncio,
        status: PNStatus,
    ) -> None:
        """Handler called on status updates."""
        operation = PubNubOperationType(status.operation).name
        category = PubNubStatusCategory(status.category).name
        if status.is_error():
            _LOGGER.error(
                "PubNub status error - operation: %s, category: %s, error: %s",
                operation,
                category,
                status.error_data.information,
            )
        else:
            _LOGGER.info(
                "PubNub status update - operation: %s, category: %s",
                operation,
                category,
            )

    def message(
        self,
        pubnub: pubnub_asyncio.PubNubAsyncio,
        message: PNMessageResult,
    ) -> None:
        """Handler called on each message received."""
        self.__message_received_callback(message.message)


@unique
class PubNubStatusCategory(IntEnum):
    UNKNOWN = 1
    ACKNOWLEDGMENT = 2
    ACCESS_DENIED = 3
    TIMEOUT = 4
    NETWORK_ISSUES = 5
    CONNECTED = 6
    RECONNECTED = 7
    DISCONNECTED = 8
    UNEXPECTED_DISCONNECT = 9
    CANCELLED = 10
    BAD_REQUEST = 11
    MALFORMED_FILTER_EXPRESSION = 12
    MALFORMED_RESPONSE = 13
    DECRYPTION_ERROR = 14
    TLS_CONNECTION_FAILED = 15
    TLS_UNTRUSTED_CERTIFICATE = 16
    INTERNAL_EXCEPTION = 17

    # Handle unknown/future operation types.
    UNEXPECTED_STATUS_CATEGORY = -1

    @classmethod
    def _missing_(cls, _):
        _LOGGER.error("Unexpected status category: %s", _)
        return cls.UNEXPECTED_STATUS_CATEGORY


@unique
class PubNubOperationType(IntEnum):
    SUBSCRIBE = 1
    UNSUBSCRIBE = 2
    PUBLISH = 3
    HISTORY = 4
    WHERE_NOW = 5

    HEARTBEAT = 6
    SET_STATE = 7
    ADD_CHANNELS_TO_GROUP = 8
    REMOVE_CHANNELS_FROM_GROUP = 9
    CHANNEL_GROUPS = 10
    REMOVE_GROUP = 11
    CHANNELS_FOR_GROUP = 12
    PUSH_NOTIFICATION_ENABLED_CHANNELS = 13
    ADD_PUSH_NOTIFICATIONS_ON_CHANNELS = 14
    REMOVE_PUSH_NOTIFICATIONS_FROM_CHANNELS = 15
    REMOVE_ALL_PUSH_NOTIFICATIONS = 16
    TIME = 17

    HERE_NOW = 18
    GET_STATE = 19
    ACCESS_MANAGER_AUDIT = 20
    ACCESS_MANAGER_GRANT = 21
    ACCESS_MANAGER_REVOKE = 22
    HISTORY_DELETE = 23
    MESSAGE_COUNT = 24
    FIRE = 25
    SIGNAL = 26

    ACCESS_MANAGER_GRANT_TOKEN = 41
    ADD_MESSAGE_ACTION = 42
    GET_MESSAGE_ACTIONS = 43
    DELETE_MESSAGE_ACTION = 44
    FETCH_MESSAGES = 45

    GET_FILES_ACTION = 46
    DELETE_FILE = 47
    GET_FILE_DOWNLOAD_URL_ACTION = 48
    FETCH_FILE_UPLOAD_S3_DATA_ACTION = 49
    DOWNLOAD_FILE_ACTION = 50
    SEND_FILE_ACTION = 51
    SEND_FILE_NOTIFICATION = 52

    SET_UUID_METADATA = 53
    GET_UUID_METADATA = 54
    REMOVE_UUID_METADATA = 55
    GET_ALL_UUID_METADATA = 56

    SET_CHANNEL_METADATA = 57
    GET_CHANNEL_METADATA = 58
    REMOVE_CHANNEL_METADATA = 59
    GET_ALL_CHANNEL_METADATA = 60

    SET_CHANNEL_MEMBERS = 61
    GET_CHANNEL_MEMBERS = 62
    REMOVE_CHANNEL_MEMBERS = 63
    MANAGE_CHANNEL_MEMBERS = 64

    SET_MEMBERSHIPS = 65
    GET_MEMBERSHIPS = 66
    REMOVE_MEMBERSHIPS = 67
    MANAGE_MEMBERSHIPS = 68

    # Handle unknown/future operation types.
    UNEXPECTED_OPERATION_TYPE = -1

    @classmethod
    def _missing_(cls, _):
        _LOGGER.error("Unexpected operation type: %s", _)
        return cls.UNEXPECTED_OPERATION_TYPE



"""Module that implements the VivintSkyApi class."""
import json
import logging
import re
import ssl
from datetime import datetime
from types import MethodType
from typing import Any, Dict, Optional

import aiohttp
import certifi
from aiohttp.client_reqrep import ClientResponse

from .const import SwitchAttribute, VivintDeviceAttribute, WirelessSensorAttribute
from .enums import ArmedState, GarageDoorState, ZoneBypass
from .exceptions import VivintSkyApiAuthenticationError, VivintSkyApiError

_LOGGER = logging.getLogger(__name__)

VIVINT_API_ENDPOINT = "https://www.vivintsky.com/api"


class VivintSkyApi:
    """Class to communicate with the VivintSky API."""

    def __init__(
        self,
        username: str,
        password: str,
        client_session: Optional[aiohttp.ClientSession] = None,
    ):
        self.__username = username
        self.__password = password
        self.__client_session = client_session or self.__get_new_client_session()
        self.__has_custom_client_session = client_session is not None
        self.__zwave_device_info = {}

    def is_session_valid(self) -> dict:
        """Return the state of the current session."""
        cookie = self.__client_session.cookie_jar._cookies["www.vivintsky.com"].get("s")
        if not cookie:
            return False
        cookie_expiration = datetime.strptime(
            cookie.get("expires"), "%a, %d %b %Y %H:%M:%S %Z"
        )
        return True if cookie_expiration > datetime.utcnow() else False

    async def connect(self) -> dict:
        """Connect to VivintSky Cloud Service."""
        authuser_data = await self.__get_vivintsky_session(
            self.__username, self.__password
        )
        if not authuser_data:
            raise VivintSkyApiAuthenticationError("Unable to login to Vivint.")
        return authuser_data

    async def disconnect(self) -> None:
        """Disconnect from VivintSky Cloud Service."""
        if not self.__has_custom_client_session:
            await self.__client_session.close()

    async def get_authuser_data(self) -> dict:
        """
        Get the authuser data

        Poll the Vivint authuser API endpoint resource to gather user-related data including enumeration of the systems
        that user has access to.
        """
        resp = await self.__get("authuser")
        async with resp:
            if resp.status == 200:
                return await resp.json(encoding="utf-8")
            else:
                raise VivintSkyApiAuthenticationError("Missing auth user data.")

    async def get_panel_credentials(self, panel_id: int) -> dict:
        """Get the panel credentials."""
        resp = await self.__get(f"panel-login/{panel_id}")
        async with resp:
            if resp.status == 200:
                return await resp.json(encoding="utf-8")
            else:
                raise VivintSkyApiAuthenticationError(
                    "Unable to retrieve panel credentials."
                )

    async def get_system_data(self, panel_id: int) -> dict:
        """Gets the raw data for a system."""
        resp = await self.__get(
            f"systems/{panel_id}",
            headers={"Accept-Encoding": "application/json"},
            params={"includerules": "false"},
        )
        async with resp:
            if resp.status == 200:
                return await resp.json(encoding="utf-8")
            else:
                raise VivintSkyApiError("Unable to retrieve system data.")

    async def get_device_data(self, panel_id: int, device_id: int) -> dict:
        """Gets the raw data for a device."""
        resp = await self.__get(
            f"system/{panel_id}/device/{device_id}",
            headers={"Accept-Encoding": "application/json"},
        )
        async with resp:
            if resp.status == 200:
                return await resp.json(encoding="utf-8")
            else:
                raise VivintSkyApiError("Unable to retrieve device data.")

    async def set_alarm_state(
        self, panel_id: int, partition_id: int, state: bool
    ) -> aiohttp.ClientResponse:
        resp = await self.__put(
            f"{panel_id}/{partition_id}/armedstates",
            headers={"Content-Type": "application/json;charset=UTF-8"},
            data=json.dumps(
                {
                    "system": panel_id,
                    "partitionId": partition_id,
                    "armState": state,
                    "forceArm": False,
                }
            ).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                resp_body = await resp.text()
                _LOGGER.error(
                    f"failed to set state {ArmedState.name(state)}. Code: {resp.status}, body: {resp_body},"
                    f"request url: {resp.request_info}"
                )
                raise VivintSkyApiError(
                    f"failed to set alarm status {ArmedState.name(state)} for panel {self.id}"
                )

    async def set_garage_door_state(
        self, panel_id: int, partition_id: int, device_id: int, state: int
    ) -> None:
        """Open/Close garage door."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/door/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: state,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(
                    f"failed to set state to {GarageDoorState.name(state)} for garage door: {device_id} @ {panel_id}:{partition_id}"
                )
                raise VivintSkyApiError(f"failed to update garage door state")

    async def set_lock_state(
        self, panel_id: int, partition_id: int, device_id: int, locked: bool
    ) -> None:
        """Lock/Unlock door lock."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/locks/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: locked,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(
                    f"failed to set status locked: {locked} for lock: {device_id} @ {panel_id}:{partition_id}"
                )
                raise VivintSkyApiError(f"failed to update lock status")

    async def set_sensor_state(
        self, panel_id: int, partition_id: int, device_id: int, bypass: bool
    ) -> None:
        """Bypass/unbypass a sensor."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/sensors/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    WirelessSensorAttribute.BYPASSED: ZoneBypass.MANUALLY_BYPASSED
                    if bypass
                    else ZoneBypass.UNBYPASSED,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(
                    f"Failed to set bypass: {bypass} for sensor: {device_id} @ {panel_id}:{partition_id}"
                )
                raise VivintSkyApiError(f"Failed to update sensor status.")

    async def set_switch_state(
        self,
        panel_id: int,
        partition_id: int,
        device_id: int,
        on: Optional[bool] = None,
        level: Optional[int] = None,
    ) -> None:
        """Set switch state."""
        # validate input
        if on is None and level is None:
            raise VivintSkyApiError('Either "on" or "level" must be provided.')
        elif level and (0 > level or level > 100):
            raise VivintSkyApiError('The value for "level" must be between 0 and 100.')

        resp = await self.__put(
            f"{panel_id}/{partition_id}/switches/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    SwitchAttribute.ID: device_id,
                    **(
                        {SwitchAttribute.STATE: on}
                        if level is None
                        else {SwitchAttribute.VALUE: level}
                    ),
                }
            ).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(
                    f"Failed to set {'on' if level is None else 'level'} to {on if level is None else level} for switch: {device_id} @ {panel_id}:{partition_id}."
                )
                raise VivintSkyApiError("Failed to update switch state.")

    async def set_thermostat_state(
        self, panel_id: int, partition_id: int, device_id: int, **kwargs
    ) -> None:
        """Set thermostat state."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/thermostats/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(kwargs).encode("utf-8"),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(
                    "Failed to set state to: %s for thermostat: %s @ %s:%s",
                    kwargs,
                    device_id,
                    panel_id,
                    partition_id,
                )
                raise VivintSkyApiError(f"Failed to update thermostat state")

    async def request_camera_thumbnail(
        self, panel_id: int, partition_id: int, device_id: int
    ) -> None:
        resp = await self.__get(
            f"{panel_id}/{partition_id}/{device_id}/request-camera-thumbnail",
        )
        async with resp:
            if resp.status < 200 or resp.status > 299:
                _LOGGER.info(
                    f"failed to request thumbnail for camera id {self.id}. Error code: {resp.status}"
                )

    async def get_camera_thumbnail_url(
        self,
        panel_id: int,
        partition_id: int,
        device_id: int,
        thumbnail_timestamp: datetime,
    ) -> str:
        resp = await self.__get(
            f"{panel_id}/{partition_id}/{device_id}/camera-thumbnail",
            params={"time": thumbnail_timestamp},
            allow_redirects=False,
        )
        async with resp:
            if resp.status != 302:
                _LOGGER.info(
                    f"failed to request thumbnail for camera id {self.id}. Status code: {resp.status}"
                )
                return

            return resp.headers.get("Location")

    def __get_new_client_session(self) -> aiohttp.ClientSession:
        """Create a new aiohttp.ClientSession object."""
        ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
        )
        connector = aiohttp.TCPConnector(enable_cleanup_closed=True, ssl=ssl_context)

        return aiohttp.ClientSession(connector=connector)

    async def __get_vivintsky_session(self, username: str, password: str) -> dict:
        """Login into the Vivint Sky platform with the given username and password.

        Returns auth user data if successful.
        """
        resp = await self.__post(
            "login",
            data=json.dumps({"username": username, "password": password}).encode(
                "utf-8"
            ),
        )
        async with resp:
            data = await resp.json(encoding="utf-8")
            if resp.status == 200:
                return data
            elif resp.status == 401:
                raise VivintSkyApiAuthenticationError(data["msg"])
            else:
                resp.raise_for_status()
                return None

    async def __get(
        self,
        path: str,
        headers: Dict[str, Any] = None,
        params: Dict[str, Any] = None,
        allow_redirects: bool = None,
    ) -> ClientResponse:
        """Perform a get request."""
        return await self.__call(
            self.__client_session.get,
            path,
            headers=headers,
            params=params,
            allow_redirects=allow_redirects,
        )

    async def __post(
        self,
        path: str,
        data: bytes = None,
    ) -> ClientResponse:
        """Perform a post request."""
        return await self.__call(self.__client_session.post, path, data=data)

    async def __put(
        self,
        path: str,
        headers: Dict[str, Any] = None,
        data: bytes = None,
    ) -> ClientResponse:
        """Perform a put request."""
        return await self.__call(
            self.__client_session.put, path, headers=headers, data=data
        )

    async def __call(
        self,
        method: MethodType,
        path: str,
        headers: Dict[str, Any] = None,
        params: Dict[str, Any] = None,
        data: bytes = None,
        allow_redirects: bool = None,
    ) -> ClientResponse:
        """Perform a request with supplied parameters and reauthenticate if necessary."""
        if path != "login" and not self.is_session_valid():
            await self.connect()

        if self.__client_session.closed:
            raise VivintSkyApiError("The client session has been closed.")

        return await method(
            f"{VIVINT_API_ENDPOINT}/{path}",
            headers=headers,
            params=params,
            data=data,
            allow_redirects=allow_redirects,
        )

    async def get_zwave_details(self, manufacturer_id, product_id, product_type_id):
        """Gets the zwave details by looking up the details on the openzwave device database."""
        UNKNOWN_RESULT = ["Unknown", "Unknown"]

        # zwave_lookup = f"{manufacturer_id}:{product_id}:{product_type_id}"
        # device_info = self.__zwave_device_info.get(zwave_lookup)
        # if device_info is not None:
        #     return device_info

        # async with aiohttp.ClientSession() as session:
        #     async with session.get(
        #         url=f"http://openzwave.net/device-database/{zwave_lookup}"
        #     ) as response:
        #         if response.status == 200:
        #             text = await response.text()
        #             title = re.search("<title>(.*)</title>", text, re.IGNORECASE)[1]
        #             result = self.__zwave_device_info[zwave_lookup] = (
        #                 UNKNOWN_RESULT
        #                 if title == "Device Database"
        #                 else title.split(" - ")
        #             )
        #             return result
        #         else:
        #             response.raise_for_status()
        #             return UNKNOWN_RESULT
        return UNKNOWN_RESULT



"""Module that implements the System class."""
import logging
from typing import List

from .const import PubNubMessageAttribute, SystemAttribute
from .devices.alarm_panel import AlarmPanel
from .entity import Entity
from .utils import first_or_none
from .vivintskyapi import VivintSkyApi

_LOGGER = logging.getLogger(__name__)


class System(Entity):
    """Describe a vivint system."""

    def __init__(self, name: str, data: dict, vivintskyapi: VivintSkyApi):
        super().__init__(data)
        self.__name = name
        self.vivintskyapi = vivintskyapi
        self.alarm_panels: List[AlarmPanel] = [
            AlarmPanel(panel_data, self)
            for panel_data in self.data[SystemAttribute.SYSTEM][
                SystemAttribute.PARTITION
            ]
        ]

    @property
    def id(self) -> str:
        """System's id"""
        return self.data[SystemAttribute.SYSTEM][SystemAttribute.PANEL_ID]

    @property
    def name(self) -> str:
        """System's name"""
        return self.__name

    async def refresh(self) -> None:
        """Reloads system's data from VivintSky API."""
        system_data = await self.vivintskyapi.get_system_data(self.id)

        for panel_data in system_data[SystemAttribute.SYSTEM][
            SystemAttribute.PARTITION
        ]:
            alarm_panel = first_or_none(
                self.alarm_panels,
                lambda panel: panel.id == panel_data[SystemAttribute.PANEL_ID]
                and panel.partition_id == panel_data[SystemAttribute.PARTITION_ID],
            )
            if alarm_panel:
                alarm_panel.refresh(panel_data)
            else:
                self.alarm_panels.append(AlarmPanel(panel_data, self))

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""

        if message[PubNubMessageAttribute.TYPE] == "account_system":
            # this is a system message
            operation = message.get(PubNubMessageAttribute.OPERATION)
            data = message.get(PubNubMessageAttribute.DATA)

            if data and operation == "u":
                self.update_data(data)

        elif message[PubNubMessageAttribute.TYPE] == "account_partition":
            # this is a message for one of the devices attached to this system
            partition_id = message.get(PubNubMessageAttribute.PARTITION_ID)
            if not partition_id:
                _LOGGER.debug(
                    f"ignoring account_partition message. No partition_id specified for system {self.id}"
                )
                return

            alarm_panel = first_or_none(
                self.alarm_panels,
                lambda panel: panel.id == self.id
                and panel.partition_id == partition_id,
            )

            if not alarm_panel:
                _LOGGER.debug(
                    f"no alarm panel found for system {self.id}, partition {partition_id}"
                )
                return

            alarm_panel.handle_pubnub_message(message)



"""Module that implements the WirelessSensor class."""
import logging

from ..const import WirelessSensorAttribute as Attributes
from ..enums import EquipmentCode, EquipmentType, SensorType, ZoneBypass
from . import VivintDevice

_LOGGER = logging.getLogger(__name__)


class WirelessSensor(VivintDevice):
    """Represents a Vivint wireless sensor device."""

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}|{self.equipment_type} {self.id}, {self.name}>"
        )

    @property
    def model(self) -> str:
        """Return the equipment_code as the model of this sensor."""
        return self.equipment_code.name

    @property
    def software_version(self) -> str:
        """Return the software version of this device, if any."""
        return self.data.get(Attributes.SENSOR_FIRMWARE_VERSION) or None

    @property
    def battery_level(self) -> int:
        """Sensor's battery level."""
        battery_level = self.data.get(Attributes.BATTERY_LEVEL)
        return (
            battery_level
            if battery_level is not None
            else 0
            if self.low_battery
            else 100
        )

    @property
    def equipment_code(self):
        """Return the equipment code of this sensor."""
        return EquipmentCode(self.data.get(Attributes.EQUIPMENT_CODE))

    @property
    def equipment_type(self):
        """Return the equipment type of this sensor."""
        return EquipmentType(self.data.get(Attributes.EQUIPMENT_TYPE))

    @property
    def sensor_type(self):
        """Return the sensor type of this sensor."""
        return SensorType(self.data.get(Attributes.SENSOR_TYPE))

    @property
    def is_bypassed(self) -> bool:
        """Return True if the sensor is bypassed."""
        return self.data[Attributes.BYPASSED] != ZoneBypass.UNBYPASSED

    @property
    def is_on(self) -> bool:
        """Return True if the sensor's state is on."""
        return self.data[Attributes.STATE]

    @property
    def low_battery(self) -> bool:
        """Return true if battery's level is low."""
        return self.data[Attributes.LOW_BATTERY]

    async def set_bypass(self, bypass: bool) -> None:
        """Bypass/unbypass the sensor."""
        await self.vivintskyapi.set_sensor_state(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, bypass
        )

    async def bypass(self) -> None:
        """Bypass the sensor."""
        await self.set_bypass(True)

    async def unbypass(self) -> None:
        """Unbypass the sensor."""
        await self.set_bypass(False)



"""Module that implements the Thermostat class."""
import logging

from ..const import ThermostatAttribute as Attribute
from ..enums import FanMode, HoldMode, OperatingMode
from . import VivintDevice
from .alarm_panel import AlarmPanel

_LOGGER = logging.getLogger(__name__)

NEST_DEVICE = "pod_nest_thermostat_device"


class Thermostat(VivintDevice):
    """Represents a Vivint thermostat device."""

    def __init__(self, data: dict, alarm_panel: AlarmPanel):
        """Initialize a thermostat."""
        super().__init__(data, alarm_panel)

        if data.get(Attribute.ACTUAL_TYPE) == NEST_DEVICE:
            [self._manufacturer, self._model] = ["Google", "Nest"]

    @property
    def cool_set_point(self) -> float:
        return self.data.get(Attribute.COOL_SET_POINT)

    @property
    def fan_mode(self) -> FanMode:
        return FanMode(self.data.get(Attribute.FAN_MODE))

    @property
    def heat_set_point(self) -> float:
        return self.data.get(Attribute.HEAT_SET_POINT)

    @property
    def hold_mode(self) -> HoldMode:
        return HoldMode(self.data.get(Attribute.HOLD_MODE))

    @property
    def humidity(self) -> int:
        return self.data.get(Attribute.HUMIDITY)

    @property
    def is_fan_on(self) -> bool:
        return self.data.get(Attribute.FAN_STATE) == 1

    @property
    def is_on(self) -> int:
        return self.data.get(Attribute.OPERATING_STATE) == 1

    @property
    def maximum_temperature(self) -> float:
        return self.data.get(Attribute.MAXIMUM_TEMPERATURE)

    @property
    def minimum_temperature(self) -> float:
        return self.data.get(Attribute.MINIMUM_TEMPERATURE)

    @property
    def operating_mode(self) -> OperatingMode:
        return OperatingMode(self.data.get(Attribute.OPERATING_MODE))

    @property
    def temperature(self) -> float:
        return self.data.get(Attribute.CURRENT_TEMPERATURE)

    @staticmethod
    def celsius_to_fahrenheit(celsius: float) -> int:
        """Converts Celsius to Fahrenheit"""
        return round(celsius * 1.8 + 32)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message directed to this entity."""
        super().handle_pubnub_message(message)

        _LOGGER.debug(f"{self.name} updated")

    async def set_state(self, **kwargs) -> None:
        """Set the state of the thermostat."""
        await self.vivintskyapi.set_thermostat_state(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, **kwargs
        )



"""Module that defines various enums."""
from enum import Enum, IntEnum, unique


@unique
class ArmedState(IntEnum):
    """Armed state."""

    DISARMED = 0
    ARMING_AWAY_IN_EXIT_DELAY = 1
    ARMING_STAY_IN_EXIT_DELAY = 2
    ARMED_STAY = 3
    ARMED_AWAY = 4
    ARMED_STAY_IN_ENTRY_DELAY = 5
    ARMED_AWAY_IN_ENTRY_DELAY = 6
    ALARM = 7
    ALARM_FIRE = 8
    DISABLED = 11
    WALK_TEST = 12


@unique
class CapabilityCategoryType(IntEnum):
    """Capability category type."""

    CAMERA = 1
    DOORBELL = 2
    SWITCH = 3
    THERMOSTAT = 4
    ZWAVE = 5
    TOUCHLINK = 6
    DOOR_LOCK = 7
    MOBILE_BLACKLIST = 8
    GARAGE_DOOR = 9

    # Handle unknown/future capability categories
    UNKNOWN = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class CapabilityType(IntEnum):
    """Capability type."""

    ANIMAL_DETECTION = 42
    BINARY_ON_OFF = 7
    CAMERA_DVR_CAPABLE = 72
    CANNOT_REPORT_STATUS = 33
    CAN_CHIME = 4
    CAN_INITIATE_CALL = 66
    CAN_LOCK_UNLOCK = 30
    CHIME_EXTENDER = 36
    CLIP_CAPTURE = 3
    DELETE_ALL_EVENTS = 63
    DETER = 45
    DETER_LIGHT = 52
    DETER_LINGER_DURATION = 53
    DETER_SCHEDULE = 50
    DETER_TONE = 51
    DETER_ZONE = 49
    DIMMABLE = 8
    DOORBELL_CHIME_SELECTABLE = 56
    DOOR_STATE = 71
    FAN120_MINUTE = 23
    FAN15_MINUTE = 18
    FAN240_MINUTE = 24
    FAN30_MINUTE = 19
    FAN45_MINUTE = 20
    FAN480_MINUTE = 25
    FAN60_MINUTE = 21
    FAN960_MINUTE = 26
    HAS_MICROPHONE = 46
    HSB_LIGHTING = 9
    HUMIDITY = 64
    HUMIDITY_CONTROL = 65
    IN_HOME_CHIME_VOLUME = 54
    LIVE_VIDEO = 6
    LURKER_DETECTION = 38
    MAINTAIN_ZOOM = 48
    MAXIMUM_TEMPERATURE = 12
    MAX_COOL_LOCK_SET_POINT = 69
    MAX_HEAT_LOCK_SET_POINT = 67
    MINIMUM_TEMPERATURE = 11
    MIN_COOL_LOCK_SET_POINT = 70
    MIN_HEAT_LOCK_SET_POINT = 68
    MIN_SETPOINT_DIFFERENCE_C = 13
    MIN_SETPOINT_DIFFERENCE_F = 14
    MOBILE_BLACKLISTABLE = 31
    MOTION_DETECTION = 1
    MUTE_CHIME = 55
    NIGHT_VISION = 60
    PACKAGE_DETECTION = 39
    PACKAGE_MOVE_DETECTION = 40
    PERSON_DETECTION = 16
    PINCH_TO_ZOOM = 32
    POLYGON_ROI = 44
    PREVIEW_CHIME_IN_HOME = 58
    PRIVACY_MODE = 15
    QUIET_MODE = 17
    REBOOT_CAMERA = 62
    RECT_ROI = 47
    RESTORE_DEFAULTS = 61
    ROTATE_IMAGE = 34
    SECURITY_2_0 = 27
    SET_POINT_LOCK = 66
    SIREN_EXTENDER = 37
    SOFT_DIP_SWITCH = 29
    STATUS_LIGHT_TOGGLE = 35
    TEMPERATURE_DETECTION = 10
    TWO_WAY_AUDIO = 5
    VEHICLE_DETECTION = 41
    VIDEO_QUALITY = 59
    VISITOR_CHIME_SELECTABLE = 57
    VISITOR_DETECTION = 2
    WARPED = 43
    WI_FI = 28
    ZWAVE_DIAGNOSTICS = 22

    # Handle unknown/future capabilities
    UNKNOWN = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@unique
class DeviceType(Enum):
    """Vivint supported device types."""

    BINARY_SWITCH = "binary_switch_device"
    CAMERA = "camera_device"
    DOOR_LOCK = "door_lock_device"
    GARAGE_DOOR = "garage_door_device"
    MULTILEVEL_SWITCH = "multilevel_switch_device"
    THERMOSTAT = "thermostat_device"
    TOUCH_PANEL = "primary_touch_link_device"
    WIRELESS_SENSOR = "wireless_sensor"
    UNKNOWN = None

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@unique
class EquipmentCode(IntEnum):
    """Equipment code."""

    APOLLO_COMBO_CO = 1422
    APOLLO_COMBO_SMOKE = 1322
    CARBON_MONOXIDE_DETECTOR_345_MHZ = 1254
    CO1_CO = 860
    CO1_CO_CANADA = 859
    CO3_2_GIG_CO = 1026
    DBELL1_2_GIG_DOORBELL = 1063
    DW10_THIN_DOOR_WINDOW = 862
    DW11_THIN_DOOR_WINDOW = 1251
    DW20_RECESSED_DOOR = 863
    DW21_R_RECESSED_DOOR = 1252
    EXISTING_CO = 692
    EXISTING_DOOR_WINDOW_CONTACT = 655
    EXISTING_FLOOD_TEMP = 556
    EXISTING_GLASS_BREAK = 475
    EXISTING_HEAT = 708
    EXISTING_KEY_FOB_REMOTE = 577
    EXISTING_MOTION_DETECTOR = 609
    EXISTING_SMOKE = 616
    FIREFIGHTER_AUDIO_DETECTOR = 1269
    GARAGE01_RESOLUTION_TILT = 1061
    GB1_GLASS_BREAK = 864
    GB2_GLASS_BREAK = 1248
    HW_DW_5816 = 673
    HW_FLOOD_SENSOR_5821 = 624
    HW_GLASS_BREAK_5853 = 519
    HW_HEAT_SENSOR_5809 = 557
    HW_PANIC_PENDANT_5802_MN2 = 491
    HW_PIR_5890 = 533
    HW_PIR_5894_PI = 530
    HW_R_DW_5818_MNL = 470
    HW_SMOKE_5808_W3 = 589
    KEY1_345_4_BUTTON_KEY_FOB_REMOTE = 866
    OTHER = 0
    PAD1_345_WIRELESS_KEYPAD = 867
    PANIC1 = 868
    PANIC2 = 1253
    PIR1_MOTION = 869
    PIR2_MOTION = 1249
    RE219_FLOOD_SENSOR = 1128
    RE220_T_2_GIG_REPEATER = 1144
    RE224_DT_DSC_TRANSLATOR = 1208
    RE224_GT_GE_TRANSLATOR = 941
    RE508_X_REPEATER = 2832
    RE524_X_WIRELESS_TAKEOVER = 2830
    REPEATER_345_MHZ = 2081
    SECURE_KEY_345_MHZ = 1250
    SMKE1_SMOKE = 872
    SMKE1_SMOKE_CANADA = 871
    SMKT2_GE_SMOKE_HEAT = 895
    SMKT3_2_GIG = 1058
    SMKT6_2_GIG = 1066
    SWS1_SMART_WATER_SENSOR = 1264
    TAKE_TAKEOVER = 873
    TILT_SENSOR_2_GIG_345 = 2831
    VS_CO3_DETECTOR = 1266
    VS_SMKT_SMOKE_DETECTOR = 1267

    @classmethod
    def _missing_(cls, value):
        return cls.OTHER


@unique
class EquipmentType(IntEnum):
    """Equipment type."""

    CONTACT = 1
    EMERGENCY = 11
    FREEZE = 6
    MOTION = 2
    TEMPERATURE = 10
    WATER = 8
    UNKNOWN = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@unique
class FanMode(IntEnum):
    """Fan mode."""

    AUTO_LOW = 0
    ON_LOW = 1
    AUTO_HIGH = 2
    ON_HIGH = 3
    TIMER_15 = 99
    TIMER_30 = 100
    TIMER_60 = 101
    TIMER_45 = 102
    TIMER_120 = 103
    TIMER_240 = 104
    TIMER_480 = 105
    TIMER_960 = 106


@unique
class GarageDoorState(IntEnum):
    """Garage door state."""

    UNKNOWN = 0
    CLOSED = 1
    CLOSING = 2
    STOPPED = 3
    OPENING = 4
    OPENED = 5


@unique
class HoldMode(IntEnum):
    """Hold mode."""

    BY_SCHEDULE = 0
    UNTIL_NEXT = 1
    TWO_HOURS = 2
    PERMANENT = 3


@unique
class OperatingMode(IntEnum):
    """Operating mode."""

    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    EMERGENCY_HEAT = 4
    RESUME = 5
    FAN_ONLY = 6
    FURNACE = 7
    DRY_AIR = 8
    MOIST_AIR = 9
    AUTO_CHANGEOVER = 10
    ENERGY_SAVE_HEAT = 11
    ENERGY_SAVE_COOL = 12
    AWAY = 13
    ECO = 100


@unique
class SensorType(IntEnum):
    """Sensor type."""

    AUDIBLE_ALARM = 7
    AUXILIARY_ALARM = 8
    CARBON_MONOXIDE = 14
    DAY_ZONE = 5
    EXIT_ENTRY_1 = 1
    EXIT_ENTRY_2 = 2
    FIRE = 9
    FIRE_WITH_VERIFICATION = 16
    INTERIOR_FOLLOWER = 4
    INTERIOR_WITH_DELAY = 10
    NO_RESPONSE = 23
    PERIMETER = 3
    REPEATER = 25
    SILENT_ALARM = 6
    SILENT_BURGLARY = 24
    UNUSED = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNUSED


@unique
class ZoneBypass(IntEnum):
    """Zone bypass statuses."""

    UNBYPASSED = 0
    FORCE_BYPASSED = 1
    MANUALLY_BYPASSED = 2



"""Module that implements the AlarmPanel class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Set, Type

from ..const import (
    AlarmPanelAttribute,
    PubNubMessageAttribute,
    PubNubOperatorAttribute,
    SystemAttribute,
)
from ..enums import ArmedState, DeviceType
from ..exceptions import VivintSkyApiError
from ..utils import add_async_job, first_or_none
from ..vivintskyapi import VivintSkyApi
from . import UnknownDevice, VivintDevice, get_device_class

if TYPE_CHECKING:
    from ..system import System

_LOGGER = logging.getLogger(__name__)


class AlarmPanel(VivintDevice):
    """Describe a Vivint alarm panel."""

    def __init__(self, data: dict, system: System):
        self.system = system
        super().__init__(data)
        self.__panel_credentials = None

        # initialize devices
        self.devices: List[VivintDevice] = [
            get_device_class(device_data[AlarmPanelAttribute.TYPE])(device_data, self)
            for device_data in self.data[AlarmPanelAttribute.DEVICES]
        ]

        # store a reference to the physical panel device
        self.__panel = first_or_none(
            self.devices,
            lambda device: DeviceType(device.data.get(AlarmPanelAttribute.TYPE))
            == DeviceType.TOUCH_PANEL,
        )

    @property
    def vivintskyapi(self) -> VivintSkyApi:
        """Instance of VivitSkyApi."""
        return self.system.vivintskyapi

    @property
    def id(self) -> int:
        """Panel's id."""
        return self.data[AlarmPanelAttribute.PANEL_ID]

    @property
    def name(self) -> str:
        """Panel's name."""
        return self.system.name

    @property
    def manufacturer(self):
        """Return Vivint as the manufacturer of this panel."""
        return "Vivint"

    @property
    def model(self):
        """Return the model (panel type) of the physical panel."""
        return (
            "Sky Control"
            if self.__panel and self.__panel.data["pant"] == 1
            else "Smart Hub"
        )

    @property
    def software_version(self) -> str:
        """Return the software version of the panel."""
        return self.__panel.software_version if self.__panel else None

    @property
    def partition_id(self) -> int:
        """Panel's partition id."""
        return self.data[AlarmPanelAttribute.PARTITION_ID]

    @property
    def is_disarmed(self) -> bool:
        """Return True if alarm is disarmed."""
        return self.get_armed_state() == ArmedState.DISARMED

    @property
    def is_armed_away(self) -> bool:
        """Return True if alarm is in armed away state."""
        return self.get_armed_state() == ArmedState.ARMED_AWAY

    @property
    def is_armed_stay(self) -> bool:
        """Return True if alarm is in armed stay state."""
        return self.get_armed_state() == ArmedState.ARMED_STAY

    def get_armed_state(self):
        """Return the panel's arm state."""
        return self.data[AlarmPanelAttribute.STATE]

    async def set_armed_state(self, state: int) -> None:
        """Set the armed state for a panel."""
        _LOGGER.debug(f"setting {self.name} to {ArmedState(state).name}")
        await self.vivintskyapi.set_alarm_state(self.id, self.partition_id, state)

    async def disarm(self) -> None:
        """Disarm the alarm."""
        await self.set_armed_state(ArmedState.DISARMED)

    async def arm_stay(self) -> None:
        """Set the alarm to armed stay."""
        await self.set_armed_state(ArmedState.ARMED_STAY)

    async def arm_away(self) -> None:
        """Set the alarm to armed away."""
        await self.set_armed_state(ArmedState.ARMED_AWAY)

    async def get_panel_credentials(self) -> dict:
        """Gets the panel credentials."""
        if not self.__panel_credentials:
            self.__panel_credentials = await self.vivintskyapi.get_panel_credentials(
                self.id
            )
        return self.__panel_credentials

    def get_devices(
        self,
        device_types: Set[Type[VivintDevice]] = None,
    ) -> List[VivintDevice]:
        """Get a list of associated devices."""
        devices: List[VivintDevice] = None

        if device_types:
            devices = [
                device for device in self.devices if device.__class__ in device_types
            ]
        else:
            devices = self.devices

        return devices

    def refresh(self, data: dict, new_device: bool = False) -> None:
        """Refreshes the alarm panel."""
        if not new_device:
            self.update_data(data, override=True)
        else:
            self.data[AlarmPanelAttribute.DEVICES].extend(
                data[AlarmPanelAttribute.DEVICES]
            )

        # update associated devices
        for device_data in data[AlarmPanelAttribute.DEVICES]:
            device = first_or_none(
                self.devices,
                lambda device: device.id == device_data[AlarmPanelAttribute.ID],
            )
            if device:
                device.update_data(device_data, override=True)
            else:
                device = get_device_class(device_data[AlarmPanelAttribute.TYPE])(
                    device_data, self
                )
                self.devices.append(device)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        operation = message.get(PubNubMessageAttribute.OPERATION)
        data = message.get(PubNubMessageAttribute.DATA)
        if not data:
            _LOGGER.debug(
                f"ignoring account_partition message for panel {self.id}, partition {self.partition_id} - "
                "no data provided"
            )
            return

        if not data.get(PubNubMessageAttribute.DEVICES):
            # this message is for the panel itself
            self.update_data(data)
        else:
            # this is a message for one (or more) of the attached devices
            devices_data = data[PubNubMessageAttribute.DEVICES]

            for device_data in devices_data:
                device_id = device_data.get("_id")
                if not device_id:
                    _LOGGER.debug("No device id")
                    continue

                if operation == PubNubOperatorAttribute.CREATE:
                    add_async_job(self.handle_new_device, device_id)
                else:
                    device = first_or_none(
                        self.devices, lambda device: device.id == device_id
                    )
                    if not device:
                        _LOGGER.debug(
                            "Ignoring message for device %s: device not found",
                            device_id,
                        )
                        continue

                    # for the sake of consistency, we also need to update the panel's raw data
                    raw_device_data = first_or_none(
                        self.data[AlarmPanelAttribute.DEVICES],
                        lambda raw_device_data: raw_device_data["_id"]
                        == device_data["_id"],
                    )

                    if operation == PubNubOperatorAttribute.DELETE:
                        self.devices.remove(device)
                        self.data[AlarmPanelAttribute.DEVICES].remove(raw_device_data)
                    else:
                        device.handle_pubnub_message(device_data)
                        raw_device_data.update(device_data)

    async def handle_new_device(self, device_id: int) -> None:
        """Handle a new device."""
        try:
            resp = await self.vivintskyapi.get_device_data(self.id, device_id)
            data = resp[SystemAttribute.SYSTEM][SystemAttribute.PARTITION][0]
            self.refresh(data, new_device=True)
        except VivintSkyApiError:
            _LOGGER.error("Error getting new device data for device %s", device_id)
